from dataclasses import dataclass, field
from math import floor
from typing import Any, Dict, List, Sequence, Tuple

import numpy as np
from stan.fit import Fit  # type: ignore


@dataclass
class MCMCRun:
    """A Stan fit, plus some derived data."""

    samples: Fit

    def iter_chains(self, param: str, index: int = 0) -> np.ndarray:
        """Return (n_chains × n_samples) matrix of draws for a given parameter."""
        n_chains, n_samples = self.samples.num_chains, self.samples.num_samples
        xss = np.zeros((n_chains, n_samples))
        for n in range(0, n_chains):
            xss[n] = self.samples[param][index][n * n_samples: (n + 1) * n_samples]
        return xss

    @property
    def params(self) -> List[str]:
        return [*self.samples.param_names]


@dataclass
class Slice:
    """A slice of an MCMC trace."""

    lower: float
    upper: float


Domain = Tuple[float, float]  # normalised domain of a plot
Param = str
Slices = Dict[Param, List[Slice]]
Props = Dict[str, Any]
Point = Tuple[float, float]


def scale(factor: float, xs: Sequence[float]) -> List[float]:
    """Element-wise product."""
    return [x * factor for x in xs]


def translate(offset: float, xs: Sequence[float]) -> List[float]:
    """Element-wise addition of a constant."""
    return [x + offset for x in xs]


def segment(domain: Domain, n: int, m: int) -> Domain:
    """Break supplied "domain" into n equal-sized segments, and return the mth."""
    assert n > 0 and 0 <= m < n
    start, end = domain
    width = (end - start) / n
    return start + m * width, start + (m + 1) * width


def to_domain(x: float, domain: Domain) -> float:
    """Convert normalised coordinate to position within supplied domain."""
    start, end = domain
    return start + x * (end - start)


@dataclass
class ParameterData:
    """MCMC data for a given parameter."""

    mcmc_run: MCMCRun
    param: str
    chains: np.ndarray = field(init=False)  # shape is [n, n_iter] where n is number of chains
    max_sample: float = field(init=False)
    min_sample: float = field(init=False)

    # cache some properties which are expensive to compute
    def __post_init__(self) -> None:
        self.chains = self.mcmc_run.iter_chains(self.param)
        self.max_sample = np.amax(self.mcmc_run.samples[self.param])
        self.min_sample = np.amin(self.mcmc_run.samples[self.param])

    @property
    def n_iter(self) -> int:
        """Return number of MCMC iterations per chain."""
        return int(self.chains.shape[1])

    def variance(self, n: int, span: int) -> List[float]:
        """For chain n, the variance over the interval [-span, span], computed pointwise."""
        assert span >= 0
        return [
            np.var(self.chains[n][max(0, i - span):min(self.n_iter + 1, i + span + 1)])
            for i in range(0, self.n_iter)
        ]


@dataclass
class ParameterSlices(ParameterData):
    """Parameter data, plus a set of slices."""

    slcs: List[Slice]

    def chain_slices(self, slc: Slice) -> List[np.ndarray]:
        """The specified slice of each chain."""
        return [
            self.chains[
                n,
                floor(slc.lower * self.n_iter):floor(slc.upper * self.n_iter)
            ]
            for n, _ in enumerate(self.chains)
        ]



from __future__ import annotations

from dataclasses import dataclass
import pickle
from typing import cast

from stan.fit import Fit  # type: ignore
from stan.model import Model  # type: ignore


@dataclass
class Stan:
    """Bundle a Stan fit with its model."""

    model: Model
    fit: Fit

    def save(self, file: str) -> None:
        """Save Stan model and fit using supplied file name."""
        with open(file + ".pkl", "wb") as f:
            pickle.dump(self, f, protocol=-1)

    @staticmethod
    def load(file: str) -> Stan:
        """Load Stan model and fit using supplied file name."""
        with open(file + ".pkl", "rb") as f:
            return cast(Stan, pickle.load(f))

    def equal(self, other: Stan) -> bool:
        """For now just compare fit for (str) equality. Model name seems to vary unpredictably."""
        return str(self.fit) == str(other.fit)



from dataclasses import dataclass
from math import floor, nan, pi
from typing import List, Sequence, Tuple

import numpy as np
from plotly.basedatatypes import BaseTraceType  # type: ignore
import plotly.graph_objects as go  # type: ignore

from backfillz.data import Domain, MCMCRun, ParameterSlices, Props, segment, Slice, to_domain
from backfillz.plot import (
    AggregatePlot, alpha, Axis, background_rect, fresh_axis_id, LeafPlot, left_vertical_title, normalise,
    Plot, polar_plot, RootPlot, tick_every
)
from backfillz.slice_histograms import SliceHistogram
from backfillz.theme import BackfillzTheme


@dataclass
class DialPlot(LeafPlot[ParameterSlices]):
    """Trace dial plot (3/4 segment)."""

    radial_domain: Domain = 1 / 3, 1.0
    angular_domain: Domain = 0.5 * pi, 2 * pi

    @property
    def plot_elements(self) -> List[BaseTraceType]:
        return [*self.donut_segments, *self.polar_traces, *self.inner_ticks]

    @property
    def xaxis_props(self) -> Props:
        # scaleanchor='y', scaleratio=1 would force an exact square/circle, but then can't
        # set histogram widths correctly
        return dict(visible=False)

    @property
    def yaxis_props(self) -> Props:
        return dict(range=[-1, 1], visible=False)

    @staticmethod
    def arc(x_domain: Domain, y: float, n_segments: int) -> Tuple[List[float], List[float]]:
        xs = [*range(0, n_segments)]
        ys = [y] * n_segments
        return polar_plot(xs, ys, normalise(xs, x_domain), Axis((0.0, 1.0), DialPlot.radial_domain))

    @staticmethod
    def donut_segment(x_domain: Domain, fillcolor: str) -> go.Scatter:
        xs1, ys1 = DialPlot.arc(x_domain, 1.0, 100)
        xs2, ys2 = DialPlot.arc(x_domain, 0.0, 50)
        xs = xs1 + xs2[::-1]
        ys = ys1 + ys2[::-1]
        return go.Scatter(x=xs, y=ys, line=dict(width=0), fill='toself', fillcolor=fillcolor)

    @staticmethod
    def slice_domain(slc: Slice) -> Domain:
        return to_domain(slc.lower, DialPlot.angular_domain), to_domain(slc.upper, DialPlot.angular_domain)

    @property
    def donut_segments(self) -> List[go.Scatter]:
        [burn_in, remaining] = self.data.slcs
        colours = DerivativeColours(self.theme)
        return [
            DialPlot.donut_segment(DialPlot.slice_domain(burn_in), colours.burn_in_segment),
            DialPlot.donut_segment(DialPlot.slice_domain(remaining), colours.remaining_segment)
        ]

    @property
    def angular_axis(self) -> Axis:
        return Axis((0, self.data.n_iter), DialPlot.angular_domain)

    @property
    def radial_axis(self) -> Axis:
        return Axis((self.data.min_sample, self.data.max_sample), DialPlot.radial_domain)

    @property
    def inner_ticks(self) -> List[go.Scatter]:
        ticks_per_circle = 80  # somewhat arbitrary
        tick_gap: int = tick_every(ticks_per_circle, self.angular_axis)
        start, end = self.angular_axis.range
        xs1 = [x * tick_gap for x in range(floor(start), floor(end / tick_gap))]
        xs2 = [start, TraceDial.burn_in_iter, end]
        top, bottom1, bottom2 = -0.04, -0.07, -0.14
        return [
            self.radial_ticks(xs1, (top, bottom1), self.theme.mg_colour),
            self.radial_ticks(xs2, (top, bottom2), self.theme.fg_colour),
            self.radial_tick_marks(xs2, bottom2)
        ]

    def radial_tick_marks(self, xs: Sequence[float], tick_bottom: float) -> go.Scatter:
        y_axis: Axis = Axis((0.0, 1.0), DialPlot.radial_domain)
        x, y = polar_plot(xs, [tick_bottom - 0.05] * len(xs), self.angular_axis, y_axis)
        return go.Scatter(x=x, y=y, text=[str(x) for x in xs], mode='text', textposition='middle left')

    def radial_ticks(self, xs: Sequence[float], tick_size: Tuple[float, float], colour: str) -> go.Scatter:
        """Ticks at supplied angular positions, sized relative to radial_domain."""
        top, bottom = tick_size
        y_axis: Axis = Axis((0.0, 1.0), DialPlot.radial_domain)
        x1, y1 = polar_plot(xs, [top] * len(xs), self.angular_axis, y_axis)
        x2, y2 = polar_plot(xs, [bottom] * len(xs), self.angular_axis, y_axis)
        x = [x for p in zip(x1, x2, x2) for x in p]
        y = [y for p in zip(y1, y2, [nan] * len(x2)) for y in p]
        return go.Scatter(x=x, y=y, mode='lines', line=dict(width=1, color=colour))

    def polar_trace(self, n: int, x_axis: Axis, y_axis: Axis) -> go.Scatter:
        chain: np.ndarray = self.data.chains[n]
        xs, ys = polar_plot([*range(0, len(chain))], [*chain], x_axis, y_axis)
        return go.Scatter(x=xs, y=ys, line=dict(color=self.theme.palette[n]))

    @property
    def polar_traces(self) -> List[go.Scatter]:
        return [
            self.polar_trace(n, self.angular_axis, self.radial_axis)
            for n, _ in enumerate(self.data.chains)
        ]


@dataclass
class TraceDialHistogram(SliceHistogram):
    """Slice histogram for trace dial plot."""

    bin_size: float = 1.0

    @property
    def plot_elements(self) -> List[BaseTraceType]:
        return [
            *[self.step_plot(n) for n, _ in enumerate(self.data.chains)],
            self.histo([*range(0, len(self.data.chains))], self.theme.mg_colour, TraceDialHistogram.bin_size),
        ]

    def step_plot(self, n: int) -> go.Scatter:
        ys, xs = self.bins([n], TraceDialHistogram.bin_size)
        return go.Scatter(
            x=xs, y=ys,
            mode='lines',
            line=dict(width=1, color=self.theme.palette[n], shape='hvh'),
            xaxis='x' + self.axis_id, yaxis='y' + self.axis_id,
        )

    @property
    def xaxis_props(self) -> Props:
        props: Props
        if self.n_slc == len(self.data.slcs) - 1:
            props = dict(side='top')
        else:
            props = dict(visible=False)
        return {**props, **dict(range=(self.data.min_sample, self.data.max_sample))}


@dataclass
class SliceHistograms(AggregatePlot[ParameterSlices]):
    """One slice histogram per slice."""

    def make_plots(self) -> List[Plot[ParameterSlices]]:
        return [
            TraceDialHistogram(
                axis_id=fresh_axis_id(),
                x_domain=self.x_domain,
                y_domain=segment(self.y_domain, len(self.data.slcs), n),
                data=self.data,
                theme=self.theme,
                slc=slc,
                n_slc=n,
            )
            for n, slc in enumerate(self.data.slcs)
        ]


@dataclass
class TraceDial(RootPlot[ParameterSlices]):
    """Top-level plot, for a given parameter and chain."""

    burn_in_iter: int = 500  # hard-coded for now -- should be a parameter?

    def make_plots(self) -> List[Plot[ParameterSlices]]:
        return [self.dial_plot, self.histograms]

    @property
    def dial_plot(self) -> DialPlot:
        return DialPlot(
            axis_id='',  # Plotly default axes
            x_domain=(0.0, 1),
            y_domain=(0.0, 1),
            data=self.data,
            theme=self.theme,
        )

    @property
    def histograms(self) -> SliceHistograms:
        start, end = DialPlot.radial_domain
        return SliceHistograms(
            x_domain=(0.5 + start * 0.5, 0.5 + end * 0.5),
            y_domain=(0.5, 1.0),
            data=self.data,
            theme=self.theme,
        )

    @property
    def title(self) -> str:
        return f"Pretzel plot for {self.data.param}"

    @property
    def burn_in_histo(self) -> Plot[ParameterSlices]:
        return self.histograms.plots[0]

    @property
    def sample_histo(self) -> Plot[ParameterSlices]:
        return self.histograms.plots[1]

    @property
    def layout_props(self) -> Props:
        colours = DerivativeColours(self.theme)
        # plotting region won't be exactly square but best we can do to align histogram width with donut
        length: int = 800
        return dict(
            width=length,
            height=length,
            shapes=[
                background_rect(self.burn_in_histo, colours.burn_in_segment),
                background_rect(self.sample_histo, colours.remaining_segment)
            ]
        )

    def add_additional_titles(self, fig: go.Figure) -> None:
        left_vertical_title(fig, self.burn_in_histo, "Burn-in histogram")
        left_vertical_title(fig, self.sample_histo, "Sample histogram")

    @staticmethod
    def fig(mcmc_run: MCMCRun, theme: BackfillzTheme, verbose: bool, param: str) -> go.Figure:
        """Create a trace slice histogram."""
        burn_in_end: float = TraceDial.burn_in_iter / mcmc_run.samples.num_samples
        slcs: List[Slice] = [Slice(0.0, burn_in_end), Slice(burn_in_end, 1)]
        return TraceDial(
            x_domain=(0.0, 1.0),
            y_domain=(0.0, 1.0),
            data=ParameterSlices(mcmc_run, param, slcs),
            theme=theme,
            verbose=verbose
        ).make_fig()


@dataclass
class DerivativeColours:
    """Colours uniquely determined by a theme."""

    theme: BackfillzTheme

    @property
    def burn_in_segment(self) -> str:
        return alpha(self.theme.mg_colour, self.theme.alpha + 0.2)

    @property
    def remaining_segment(self) -> str:
        return alpha(self.theme.mg_colour, self.theme.alpha - 0.3)



from dataclasses import dataclass, field
from math import cos, floor, log10, pi, sin
from typing import Any, Dict, Generic, List, Literal, Optional, Sequence, Tuple, TypeVar

from plotly.basedatatypes import BaseTraceType  # type: ignore
from plotly.colors import hex_to_rgb  # type: ignore
import plotly.graph_objects as go  # type: ignore

from backfillz.data import Domain, ParameterData, Point, Props
from backfillz.theme import BackfillzTheme


class AbstractMethodError(NotImplementedError):
    """MyPy doesn't support abstract data classes yet (https://github.com/python/mypy/issues/5374)."""

    pass


AxisId = str
axis_count = 2  # start with 2 for consistently with Plotly

T = TypeVar('T', bound='ParameterData')


@dataclass
class Plot(Generic[T]):
    """Base class providing common subplot functionality."""

    x_domain: Domain  # left/right edges normalised to [0, 1]
    y_domain: Domain  # top/bottom edges normalised to [0, 1]
    data: T
    theme: BackfillzTheme

    def layout_axes(self, fig: go.Figure) -> None:
        raise AbstractMethodError()

    def render(self, fig: go.Figure) -> None:
        """Render me into fig."""
        raise AbstractMethodError()

    @property
    def top_left(self) -> Point:
        return self.x_domain[0], self.y_domain[1]

    @property
    def bottom_right(self) -> Point:
        return self.x_domain[1], self.y_domain[0]


@dataclass
class LeafPlot(Plot[T]):
    """A leaf subplot."""

    # Either generated using fresh_axis_id, or '' to mean the figure's default axes.
    axis_id: AxisId

    @property
    def plot_elements(self) -> Sequence[BaseTraceType]:
        raise AbstractMethodError()

    def render(self, fig: go.Figure) -> None:
        for el in self.plot_elements:
            fig.add_trace(el)

    @property
    def axis_defaults(self) -> Dict[str, Any]:
        return dict(
            showgrid=False,
            zeroline=False,
            linecolor=self.theme.fg_colour,
            ticks='outside',
            tickwidth=1,
            ticklen=5,
            tickcolor=self.theme.fg_colour,
            fixedrange=True,  # disable selection zoom
        )

    @property
    def xaxis_id(self) -> str:
        """My Plotly x-axis id."""
        return 'xaxis' + self.axis_id

    @property
    def yaxis_id(self) -> str:
        """My Plotly y-axis id."""
        return 'yaxis' + self.axis_id

    def layout_axes(self, fig: go.Figure) -> None:
        """Configure my x and y axis settings in fig."""
        fig.update_layout({
            self.xaxis_id: dict(anchor='y' + self.axis_id),
            self.yaxis_id: dict(anchor='x' + self.axis_id)
        })
        fig.layout[self.xaxis_id].update(domain=self.x_domain, **self.axis_defaults, **self.xaxis_props)
        fig.layout[self.yaxis_id].update(domain=self.y_domain, **self.axis_defaults, **self.yaxis_props)

    @property
    def xaxis_props(self) -> Props:
        """My custom x-axis settings; subclasses can override."""
        return dict()

    @property
    def yaxis_props(self) -> Props:
        """My custom y-axis settings; subclasses can override."""
        return dict()


@dataclass
class AggregatePlot(Plot[T]):
    """A collection of subplots."""

    plots: Sequence[Plot[T]] = field(init=False)

    def __post_init__(self) -> None:
        self.plots = self.make_plots()

    def make_plots(self) -> Sequence[Plot[T]]:
        """My subplots."""
        raise AbstractMethodError()

    def layout_axes(self, fig: go.Figure) -> None:
        """Ask each subplot to configure its axes."""
        for plot in self.plots:
            plot.layout_axes(fig)

    def render(self, fig: go.Figure) -> None:
        """Render subplots into fig."""
        for plot in self.plots:
            plot.render(fig)


@dataclass
class RootPlot(AggregatePlot[T]):
    """Top-level plot container."""

    verbose: bool

    @property
    def title(self) -> str:
        """Title for overall figure."""
        raise AbstractMethodError()

    def add_additional_titles(self, fig: go.Figure) -> None:
        annotate(
            fig, 14, (1, -0.03), 'right', 'top', None,  # leave room for an x-axis, if needed
            "Backfillz-py by CIM, University of Warwick and The Alan Turing Institute"
        )

    @property
    def layout_props(self) -> Props:
        return dict()

    def make_fig(self) -> go.Figure:
        """Create fig and render subplots."""
        fig: go.Figure = go.Figure(
            layout=go.Layout(
                title=self.title,
                titlefont=dict(size=30),
                plot_bgcolor='rgba(0,0,0,0)',
                paper_bgcolor='rgba(0,0,0,0)',
                showlegend=False,
                bargap=0,
                **self.layout_props,
            )
        )

        self.layout_axes(fig)
        self.add_additional_titles(fig)
        self.render(fig)
        return fig


@dataclass
class Axis:
    """Map a range into a domain."""

    range: Domain
    domain: Domain

    # Don't require that r_start <= x <= r_end.
    def translate(self, xs: Sequence[float]) -> Sequence[float]:
        r_start, r_end = self.range
        d_start, d_end = self.domain
        return [(x - r_start) / (r_end - r_start) * (d_end - d_start) + d_start for x in xs]


def normalise(xs: Sequence[float], domain: Domain) -> Axis:
    """Map a data range into a domain."""
    return Axis((min(xs), max(xs)), domain)


def fresh_axis_id() -> str:
    """Allocate an axis id that hasn't been used elsewhere."""
    global axis_count
    axis_id = axis_count
    axis_count = axis_count + 1
    return str(axis_id)


def default_config() -> Props:
    """Preferred settings for Plotly figure."""
    return dict(displayModeBar=False, showAxisDragHandles=False)


def alpha(hex_colour: str, a: float) -> str:
    """Add an alpha component to a colour represented as a hex string without an alpha component."""
    rgb: tuple[int, int, int] = hex_to_rgb(hex_colour)
    return f"rgba({rgb[0]},{rgb[1]},{rgb[2]},{a})"


def annotate(
    fig: go.Figure,
    font_size: int,
    at: Point,
    xanchor: Literal['left', 'center', 'right'],
    yanchor: Literal['top', 'middle', 'bottom'],
    y_adjust: Optional[float],  # additional normalised offet of text relative to plot
    text: str,
    textangle: int = 0,
) -> None:
    """Add an annotation to supplied figure, with supplied arguments in addition to some default settings."""
    fig.add_annotation(
        xref='paper',
        yref='paper',
        showarrow=False,
        font=dict(size=font_size),
        x=at[0],
        y=at[1] + (0 if y_adjust is None else y_adjust),
        xanchor=xanchor,
        yanchor=yanchor,
        text=text,
        textangle=textangle
    )


def left_vertical_title(fig: go.Figure, plot: Plot[T], title: str) -> None:
    """Add vertical title to left of plot."""
    annotate(fig, 14, plot.top_left, 'right', 'top', None, title, textangle=-90)


def background_rect(plot: Plot[T], fillcolor: str) -> Props:
    """Shaded background for a plot, as a Plotly shape that can be added to layout.shapes."""
    x0, y0 = plot.top_left
    x1, y1 = plot.bottom_right
    return dict(
        type='rect',
        xref='paper', yref='paper',
        x0=x0, y0=y0,
        x1=x1, y1=y1,
        fillcolor=fillcolor,
        layer='below',
        line_width=0,
    )


def tick_every(ticks_per_circle: int, angular_axis: Axis) -> int:
    """Tick gap in range units, based on desired approximate number of ticks per circle."""
    dom_start, dom_end = angular_axis.domain
    num_ticks: float = (dom_end - dom_start) / (2 * pi) * ticks_per_circle
    start, end = angular_axis.range
    tick_gap: float = (end - start) / num_ticks
    return int(round(tick_gap, -int(floor(log10(abs(tick_gap))))))  # 1 sig fig


def spiral_plot(
    xs: Sequence[float],
    ys: Sequence[float],
    x_axis: Axis,
    y_axis: Axis,
    b: float,
) -> Tuple[List[float], List[float]]:
    """Plot along arithmetic spiral r = a + b * theta, via the supplied axes. 12 o'clock = 0.5 * pi."""
    assert len(xs) == len(ys)
    thetas = x_axis.translate(xs)
    ys_radial = [y + b * theta for theta, y in zip(thetas, y_axis.translate(ys))]
    rs_thetas: List[Tuple[float, float]] = [
        (cos(theta) * y, sin(theta) * y)
        for theta, y in zip(thetas, ys_radial)
    ]
    return [r for r, _ in rs_thetas], [theta for _, theta in rs_thetas]


def polar_plot(
    xs: Sequence[float],
    ys: Sequence[float],
    x_axis: Axis,
    y_axis: Axis
) -> Tuple[List[float], List[float]]:
    """A spiral plot with b = 0."""
    return spiral_plot(xs, ys, x_axis, y_axis, 0)



from datetime import datetime
from enum import Enum
import sys
from typing import List

import plotly.graph_objects as go  # type: ignore
from stan.fit import Fit  # type: ignore

from backfillz.data import MCMCRun
from backfillz.spiral_stream import SpiralStream
from backfillz.theme import BackfillzTheme, default
from backfillz.trace_dial import TraceDial
from backfillz.trace_slice_histogram import TraceSliceHistogram


class HistoryEvent(Enum):
    """Category of event."""

    OBJECT_CREATION = 1
    SLICE_HISTOGRAM = 2
    TRACE_DIAL = 3
    SPIRAL_STREAM = 4


class HistoryEntry:
    """An entry in the Backfillz history log."""

    count = 0

    ident: int
    date: datetime
    event: HistoryEvent
    python_version: str
    saved: bool

    def __init__(self, event: HistoryEvent, saved: bool) -> None:
        """Construct a history entry."""
        self.ident = HistoryEntry.count
        HistoryEntry.count += 1
        self.date = datetime.now()
        self.event = event
        self.python_version = sys.version
        self.saved = saved


class Backfillz:
    """Represents a Backfillz user session."""

    theme: BackfillzTheme
    mcmc_run: MCMCRun
    verbose: bool
    plot_history: List[HistoryEntry]

    def __init__(self, fit: Fit, verbose: bool = False) -> None:
        """Initialise a Backfillz session."""
        self.mcmc_run = MCMCRun(fit)
        self.verbose = verbose
        self.set_theme(default)
        self.plot_history = [
            HistoryEntry(HistoryEvent.OBJECT_CREATION, False)
        ]

    def set_theme(self, theme: BackfillzTheme) -> None:
        """Set Backfillz theme."""
        if self.verbose:
            print("Setting backfillz object theme to " + theme.name)
        self.theme = theme

    def plot_slice_histogram(self, param: str, save_plot: bool = False) -> go.Figure:
        """Create and plot a slice histogram."""
        fig = TraceSliceHistogram.fig(self.mcmc_run, self.theme, self.verbose, param)
        self.plot_history.append(HistoryEntry(HistoryEvent.SLICE_HISTOGRAM, save_plot))
        return fig

    def plot_trace_dial(self, param: str, save_plot: bool = False) -> go.Figure:
        """Create and plot a trace dial."""
        fig = TraceDial.fig(self.mcmc_run, self.theme, self.verbose, param)
        self.plot_history.append(HistoryEntry(HistoryEvent.TRACE_DIAL, save_plot))
        return fig

    def plot_spiral_stream(self, param: str, steps: List[int], save_plot: bool = False) -> go.Figure:
        """Create and plot a spiral stream."""
        fig = SpiralStream.fig(self.mcmc_run, self.theme, self.verbose, param, steps)
        self.plot_history.append(HistoryEntry(HistoryEvent.SPIRAL_STREAM, save_plot))
        return fig



from dataclasses import dataclass
from math import floor, pi
from typing import List, Sequence, Tuple

import plotly.graph_objects as go  # type: ignore

from backfillz.data import Domain, MCMCRun, ParameterData, Props, segment
from backfillz.plot import (
    AggregatePlot, alpha, annotate, Axis, fresh_axis_id, LeafPlot, RootPlot, spiral_plot
)
from backfillz.theme import BackfillzTheme


@dataclass
class ParameterSteps(ParameterData):
    """Data being visualised by a spiral stream plot."""

    steps: List[int]


@dataclass
class SpiralPlot(LeafPlot[ParameterSteps]):
    """Spiral plot for chain and step."""

    n_chain: int
    step: int

    angular_domain: Domain = 0.5 * pi, 2 * pi * 3

    @property
    def plot_elements(self) -> List[go.Scatter]:
        chain: List[float] = self.data.variance(self.n_chain, self.step)
        xs: List[int] = [*range(0, len(chain))]
        x_axis: Axis = Axis((0, len(chain)), SpiralPlot.angular_domain)
        y_range: Domain = min(chain), max(chain)
        xs1, ys1 = spiral_plot(xs, chain, x_axis, Axis(y_range, (0.5, 1)), 1 / (2 * pi))
        xs2, ys2 = spiral_plot(xs, chain, x_axis, Axis(y_range, (0.5, 0)), 1 / (2 * pi))
        return [go.Scatter(
            x=xs1 + xs2[::-1],
            y=ys1 + ys2[::-1],
            fill='toself',
            fillcolor=alpha(self.theme.palette[self.n_chain], 0.5),
            line=dict(width=0.5, color=self.theme.palette[self.n_chain]),
            xaxis='x' + self.axis_id,
            yaxis='y' + self.axis_id,
        )]

    @property
    def overall_range(self) -> Tuple[float, float]:
        _, end = self.angular_domain
        rotations: int = floor(end / (2 * pi) + 1)
        return (-rotations, rotations)

    @property
    def xaxis_props(self) -> Props:
        return dict(visible=False, range=self.overall_range)

    @property
    def yaxis_props(self) -> Props:
        return dict(visible=False, range=self.overall_range)


@dataclass
class SpiralRow(AggregatePlot[ParameterSteps]):
    """Row of spiral plots for chain, one for each step."""

    n_chain: int

    def make_plots(self) -> Sequence[SpiralPlot]:
        return [
            SpiralPlot(
                data=self.data,
                theme=self.theme,
                axis_id=fresh_axis_id(),
                x_domain=segment(self.x_domain, len(self.data.steps), n),
                y_domain=self.y_domain,
                n_chain=self.n_chain,
                step=step,
            )
            for n, step in enumerate(self.data.steps)
        ]


@dataclass
class SpiralStream(RootPlot[ParameterSteps]):
    """Spiral stream plot for a given parameter; one spiral row per chain."""

    def make_plots(self) -> Sequence[SpiralRow]:
        return [
            SpiralRow(
                x_domain=(0, 1),
                y_domain=segment(self.y_domain, len(self.data.chains), n),
                data=self.data,
                theme=self.theme,
                n_chain=n,
            )
            for n, _ in enumerate(self.data.chains)
        ]

    @property
    def title(self) -> str:
        return f"Spiral stream plot for {self.data.param}"

    def add_additional_titles(self, fig: go.Figure) -> None:
        super().add_additional_titles(fig)
        annotate(fig, 14, (0.5, 0), 'center', 'middle', None, "Variance")
        annotate(fig, 14, (0, 0.5), 'right', 'middle', None, "Chain", textangle=-90)

    @property
    def layout_props(self) -> Props:
        # ensure each individual spiral plot is square; see trace_dial
        length: int = 800
        return dict(width=length, height=length * len(self.data.chains) / len(self.data.steps))

    @staticmethod
    def fig(
        mcmc_run: MCMCRun,
        theme: BackfillzTheme,
        verbose: bool,
        param: str,
        steps: List[int]
    ) -> go.Figure:
        """Create a spiral stream plot."""
        return SpiralStream(
            x_domain=(0.0, 1.0),
            y_domain=(0.0, 1.0),
            data=ParameterSteps(mcmc_run, param, steps),
            theme=theme,
            verbose=verbose,
        ).make_fig()



from dataclasses import dataclass
from math import ceil, floor
from typing import cast, List, Tuple

import numpy as np
from plotly.basedatatypes import BaseTraceType  # type: ignore
import plotly.graph_objects as go  # type: ignore
import scipy.stats as stats  # type: ignore

from backfillz.data import ParameterSlices, Props, Slice
from backfillz.plot import LeafPlot


Bins = Tuple[List[float], List[float]]


@dataclass
class SliceHistogram(LeafPlot[ParameterSlices]):
    """Histogram for arbitrary subsets of chains, plus optional KDE plots for individual chains."""

    slc: Slice
    n_slc: int

    @property
    def plot_elements(self) -> List[BaseTraceType]:
        """Histogram for a slice (aggregating subset of chains) plus density plot for each chain."""
        ns: List[int] = [*range(0, len(self.data.chains))]
        return [self.histo(ns, self.theme.fg_colour, 1), *[self.chain_plot(n) for n in ns]]

    # Histogram bins for specified subset of chains.
    def bins(self, ns: List[int], bin_size: float) -> Bins:
        return cast(Bins, np.histogram(
            [x for n in ns for x in self.data.chain_slices(self.slc)[n]],
            [*np.arange(floor(self.data.min_sample), ceil(self.data.max_sample), bin_size)],
            density=True,
        ))

    # Histogram for specified subset of chains. Compute our own bins so we're in full control.
    def histo(self, ns: List[int], color: str, bin_size: float) -> go.Bar:
        ys, xs = self.bins(ns, bin_size)
        return go.Bar(
            x=xs,
            y=ys,
            marker=dict(color=self.theme.bg_colour, line=dict(color=color, width=1)),
            xaxis='x' + self.axis_id,
            yaxis='y' + self.axis_id,
        )

    # Non-parametric KDE, smoothed with a Gaussian kernel, for a given chain.
    def chain_plot(self, n: int) -> go.Scatter:
        x = np.linspace(self.data.min_sample, self.data.max_sample, 200)
        chain_slices = self.data.chain_slices(self.slc)
        return go.Scatter(
            x=x,
            y=stats.kde.gaussian_kde(chain_slices[n])(x),
            mode='lines',
            line=dict(width=2, color=self.theme.palette[n]),
            yaxis='y' + self.axis_id,
            xaxis='x' + self.axis_id,
        )

    @property
    def xaxis_props(self) -> Props:
        bottom: bool = self.n_slc == 0
        top: bool = self.n_slc == len(self.data.slcs) - 1
        # single slice requires special treatment; haven't figured out how to mirror tick labels
        props: Props
        if len(self.data.slcs) == 1:
            props = dict(mirror='ticks')
        elif bottom:
            props = dict()
        elif top:
            props = dict(side='top')
        else:
            props = dict(visible=False)
        return {**props, **dict(range=(self.data.min_sample, self.data.max_sample))}

    @property
    def yaxis_props(self) -> Props:
        return dict(side='right', rangemode='nonnegative')
